{{#partial}}{title}{{package_name}} in {{suite}}/{{section}}{{/partial}}

{{#partial}}{ float_right }
<small>Last updated on: {{time}}</small>
{{/partial}}

{{#partial}}{ head_extra }
<link rel="stylesheet" href="{{root_url}}/static/css/highlight.css">
<script src="{{root_url}}/static/js/highlight/highlight.pack.js"></script>
{{/partial}}

{{#partial}}{ header_content }
<span style="font-size:18px;"><a href="index.html" style="color: #000000;">⇦ |</a></span>

<span>{{package_name}}</span> <span style="font-size:16px;">[{{section}}]</span>
{{/partial}}

{{#partial}}{ content }
    <h1>Metadata for {{package_name}} in {{section}}</h1>

    <div class="wrapper">
      {{#cpts}}
      <h2>
        {{component_id}}
        {{#architectures}}
          <span style="font-size:12px;" class="label label-neutral">⚙ {{arch}}</span>
        {{/architectures}}
      </h2>

      <div style="float: right; padding: 6px;">
        <img src="{{icon_url}}" alt="Icon" />
      </div>

      <div class="well">
        <pre><code>{{metadata}}</code></pre>
      </div>
      {{/cpts}}

    </div>
    <script>hljs.initHighlightingOnLoad();</script>
{{/partial}}

{{> base}}
