{{#partial}}{title}Components in {{suite}}/{{section}}{{/partial}}

{{#partial}}{ header_content }
<span style="font-size:18px;"><a href="../" style="color: #000000;">⇦ |</a></span>
<span>Components summary for {{suite}}/{{section}}</span>
{{/partial}}

{{#partial}}{ content }
    <h1>Metadata for {{suite}}/{{section}}</h1>

      {{#summaries}}
        <h3 id="{{maintainer_anchor}}">{{maintainer}} <a title="Permalink" class="permalink" href="#{{maintainer_anchor}}">#</a></h3>
        <ul class="overviewlisting">
        {{#packages}}
          <li>
            <span style="font-size:16px;">
            <a href="{{pkgname}}.html">{{pkgname}}</a>&nbsp;
            </span>

            <ul>
            {{#components}}
              <li>{{cid}}</li>
            {{/components}}
            </ul>

          </li>
        {{/packages}}
        </ul>
      {{/summaries}}

{{/partial}}

{{#partial}}{ float_right }
<small>Last updated on: {{time}}</small>
{{/partial}}

{{> base}}
