<!DOCTYPE html>
<html lang="en">
  <head>
{{#block}}{ head }
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta name="description" content="">
    <meta name="author" content="">

    <title>AppStream Report for {{project_name}} - {{#block}}{ title }{{/block}}</title>

    <link href="{{root_url}}/static/css/style.css" rel="stylesheet" type="text/css" />
    <link rel="shortcut icon" href="{{root_url}}/static/img/favicon.png">

    {{#block}}{ head_extra }{{/block}}
{{/block}}
</head>

<body>
<div class="headbar">
  <div class="headbar-content avoidwrap">
    {{#block}}{ header_content }{{/block}}
  </div>
  <div class="headbar-content-right avoidwrap">
    {{#block}}{ float_right }{{/block}}
  </div>
</div>

<div class="content">
  {{#block}}{ content }{{/block}}
</div>

<hr />

<footer>
<p style="font-size:0.8em;">
    Generated by <a href="https://github.com/ximion/appstream-generator">appstream-generator</a> (v{{generator_version}}).
</p>
{{#block}}{ page_details }{{/block}}
</footer>

</body>
</html>
